# -*- coding: utf-8 -*-

##########################
# NOTE that the tests here are only written to work with
# Python 3.6 or higher!
##########################

import datetime
import sys
import time
from collections import namedtuple
from threading import Thread

import mock
import pytest
from resilient_lib.components.poller_common import (DEFAULT_CASES_QUERY_FILTER,
                                                    IntegrationError,
                                                    SOARCommon, b_to_s,
                                                    eval_mapping,
                                                    get_last_poller_date,
                                                    poller, s_to_b)
from resilient_lib.util import constants

from resilient import SimpleHTTPException

NEW_CASE_PAYLOAD = {
  "name": "test case",
  "description": u"a short description  ฑ ฒ ณ ด ต ถ ท ธ น ",
  "discovered_date": str(datetime.datetime.now()),
  "start_date": str(datetime.datetime.now()),
  "plan_status": "A"
}



class SimplePollerTester():

    def __init__(self, interval=1):
        self.PACKAGE_NAME = "TEST_PACKAGE"
        self.polling_interval = interval
        self.last_poller_time = get_last_poller_date(10)

        self.count = 0


    @poller("polling_interval", "last_poller_time")
    def good_run(self, *args, **kwargs):
        self.count += 1
        assert "last_poller_time" in kwargs

    @poller("polling_interval", "last_poller_time")
    def bad_run(self, *args, **kwargs):
        self.count += 1
        raise Exception("some exception")

def _raise_generic_exception(*__args, **__kwargs):
    raise Exception("something went wrong!")

def _raise_http_exception(*__args, **__kwargs):
    Response = namedtuple("Response", ["reason", "text"])
    raise SimpleHTTPException(Response("No reason", "something went wrong"))

@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_poller_decorator_success():
    """
    NOTE: this test takes 'sleep_time' seconds to run as it is
    testing the poller
    """
    sleep_time = 1

    test_poller = SimplePollerTester(interval=1)

    poller_thread = Thread(target=test_poller.good_run)
    poller_thread.daemon = True
    poller_thread.start()

    time.sleep(sleep_time)
    # allow for flexibility of ± 1
    assert test_poller.count in range(sleep_time-1, sleep_time+2)

@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_poller_decorator_failure(caplog):
    """
    Raises and exception to make sure those are handled properly
    """
    sleep_time = 1

    test_poller = SimplePollerTester()

    poller_thread = Thread(target=test_poller.bad_run)
    poller_thread.daemon = True
    poller_thread.start()

    time.sleep(sleep_time)
    # allow for flexibility of ± 1
    assert test_poller.count in range(sleep_time-1, sleep_time+2)

    # the actual ERROR message has the line number, but that could easily change
    # so we're splitting up the assertions here around the line number
    # ex:
    # "ERROR    resilient_lib.components.poller_common:poller_common.py:83 some exception"
    assert "ERROR    resilient_lib.components.poller_common:poller_common.py" in caplog.text
    assert "some exception" in caplog.text



@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_get_case(fx_mock_resilient_client):
    soar_common = SOARCommon(fx_mock_resilient_client)

    case = soar_common.get_case(2314)

    assert "case_types" in case
    assert case["id"] == 2314




@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_get_soar_case(fx_mock_resilient_client):
    soar_common = SOARCommon(fx_mock_resilient_client)

    case, _err_msg = soar_common.get_soar_case({ "id": 2314 }, open_cases=False)

    assert case["plan_status"] == "A"
    assert case["id"] == 2314
    assert fx_mock_resilient_client.session.adapters.get("https://").last_request.query == DEFAULT_CASES_QUERY_FILTER




@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_get_soar_cases_success(fx_mock_resilient_client, caplog):
    soar_common = SOARCommon(fx_mock_resilient_client)

    mock_query = "return_level=notnormal"
    cases, _err_msg = soar_common.get_soar_cases({ "id": 2314, "bad_field": True }, open_cases=True, uri_filters=mock_query)

    assert cases
    assert len(cases) == 1
    assert cases[0]["id"] == 2314
    assert fx_mock_resilient_client.session.adapters.get("https://").last_request.query == mock_query

@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_get_soar_cases_failure(fx_mock_resilient_client):
    old_post = fx_mock_resilient_client.post
    fx_mock_resilient_client.post = _raise_http_exception
    soar_common = SOARCommon(fx_mock_resilient_client)

    cases, _err_msg = soar_common.get_soar_cases({ "id": 2314, "bad_field": True }, open_cases=True)

    assert "No reason:  something went wrong" in _err_msg

    # reset
    fx_mock_resilient_client.post = old_post



@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_create_soar_case_success(fx_mock_resilient_client):
    soar_common = SOARCommon(fx_mock_resilient_client)

    created_case = soar_common.create_soar_case(NEW_CASE_PAYLOAD)

    assert created_case["name"] == "test case"

@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_create_soar_case_failure(fx_mock_resilient_client):
    old_post = fx_mock_resilient_client.post
    fx_mock_resilient_client.post = _raise_generic_exception
    soar_common = SOARCommon(fx_mock_resilient_client)

    with pytest.raises(IntegrationError) as err:
        soar_common.create_soar_case(NEW_CASE_PAYLOAD)

    assert "create_soar_case failed to create case in SOAR" in str(err)

    # reset
    fx_mock_resilient_client.post = old_post




@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_update_soar_case_success(fx_mock_resilient_client):
    soar_common = SOARCommon(fx_mock_resilient_client)

    updated_case = soar_common.update_soar_case(2314, {"description": "now has a description  ฑ ฒ ณ ด ต ถ ท ธ น"})

    assert updated_case["id"] == 2314
    assert updated_case["success"]

@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_update_soar_case_failure(fx_mock_resilient_client):
    old_patch = fx_mock_resilient_client.patch
    fx_mock_resilient_client.patch = _raise_generic_exception
    soar_common = SOARCommon(fx_mock_resilient_client)

    with pytest.raises(IntegrationError) as err:
        soar_common.update_soar_case(2314, {"description": "now has a description  ฑ ฒ ณ ด ต ถ ท ธ น"})

    assert "update_soar_case failed to update case in SOAR" in str(err)

    # reset
    fx_mock_resilient_client.patch = old_patch




@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_create_case_comment(fx_mock_resilient_client):
    soar_common = SOARCommon(fx_mock_resilient_client)

    comment = "A test comment  ฑ ฒ ณ ด ต ถ ท ธ น "

    resp = soar_common.create_case_comment(2314, comment)

    assert resp["text"]["content"] == comment




@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_create_datatable_row(fx_mock_resilient_client):
    soar_common = SOARCommon(fx_mock_resilient_client)

    rowdata = {"column_1_api_name": 1, "column_2_api_name": 2}
    resp = soar_common.create_datatable_row(2314, "custom_dt", rowdata)

    assert resp["id"] == 27
    assert "cells" in resp
    assert "column_1_api_name" in resp["cells"] and "column_2_api_name" in resp["cells"]




@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_get_case_attachment(fx_mock_resilient_client):
    soar_common = SOARCommon(fx_mock_resilient_client)

    resp = soar_common.get_case_attachments(2314, return_base64=True)

    assert len(resp) == 2
    assert resp[0]["id"] == 27
    assert resp[1]["id"] == 123
    assert "content" in resp[0] and "content" in resp[1]
    assert resp[0]["content"] == "IlwiYWJjZGVmXCIi"

    resp = soar_common.get_case_attachments(2314, return_base64=False)
    assert resp[0]["content"] == b'"\\"abcdef\\""'




@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_lookup_artifact_type_failure(fx_mock_resilient_client):
    soar_common = SOARCommon(fx_mock_resilient_client)

    resp = soar_common.lookup_artifact_type("incident")

    assert not resp

@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_lookup_artifact_type_success(fx_mock_resilient_client):
    soar_common = SOARCommon(fx_mock_resilient_client)

    resp = soar_common.lookup_artifact_type("artifact")

    assert resp == "my artifact"


@pytest.mark.livetest
@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY_VERSION, reason="poller common requires python3.6 or higher")
def test_filter_soar_comments(fx_mock_resilient_client):
    soar_common = SOARCommon(fx_mock_resilient_client)

    mock_entity_comments = ["new comment to add here", "comment to add"]
    mock_soar_comments = ["new comment to add here"]
    resp = soar_common._filter_comments(mock_soar_comments, mock_entity_comments)

    assert resp == ["comment to add"]



def test_eval_mapping(caplog):
    # test convert str to dict
    mock_config = '"source":"A","tags":["tagA"],"priorities":[40,50]'
    parsed_config = eval_mapping(mock_config, "{{ {} }}")
    assert parsed_config == { "source": "A", "tags": ["tagA"], "priorities": [40, 50] }

    # test no value given
    assert not eval_mapping(None)

    # test bad result (not list or dict)
    result = eval_mapping("4, 5", "({})")
    assert not result
    assert """mapping eval_value must be a string representation of a (partial) array or dictionary e.g. "'value1', 'value2'" or "'key':'value'""" in caplog.text




def test_s_to_b_and_b_to_s():
    mocks = ["ç∆˚¬å", "test", ]
    for mock_s_val in mocks:
        b = s_to_b(mock_s_val)
        assert b

        s = b_to_s(b)
        assert s
        assert s == mock_s_val

    # check for non-decodeable/encodeable value
    assert b_to_s("1234") == "1234"
    assert s_to_b(1234) == 1234



def test_get_last_poller_date():

    with mock.patch("resilient_lib.components.poller_common._get_timestamp") as mock_time:
        mock_time.return_value = datetime.datetime.fromtimestamp(1234)

        # NOTE: these hard coded timestamps are set to work with
        # Travis's timezone so might not work locally
        x = get_last_poller_date(20)
        assert str(x) == "1970-01-01 00:00:34"


        x = get_last_poller_date(10)
        assert str(x) == "1970-01-01 00:10:34"



import json
import logging
import os
import time
import unittest

import pytest
import requests
import requests_mock
from parameterized import parameterized
from resilient_lib import (IntegrationError, RequestsCommon,
                           RequestsCommonWithoutSession)
from resilient_lib.components.requests_common import (
    get_case_insensitive_key_value, is_payload_in_json)
from tests.shared_mock_data import mock_paths

REQUESTS_COMMON_CLASSES = [[RequestsCommon], [RequestsCommonWithoutSession]]

class TestFunctionRequests(unittest.TestCase):
    """ Tests for the attachment_hash function

    live tests against endpoints are listed below
    """

    URL_TEST_DATA_RESULTS = "https://api.ipify.org"
    URL_TEST_HTTP_VERBS = "https://postman-echo.com"
    URL_TEST_HTTP_STATUS_CODES = "http://httpstat.us"
    URL_TEST_PROXY = "https://gimmeproxy.com/api/getProxy"

    #rc.execute_call(verb, url, payload, log=None, basicauth=None, verify_flag=True, headers=None,
    #            proxies=None, timeout=None, resp_type=json, callback=None):
    LOG = logging.getLogger(__name__)


    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    def test_resilient_common_proxies(self, RCObjectType):
        rc = RCObjectType()
        self.assertIsNone(rc.get_proxies())

        integrations = { }

        rc = RCObjectType(opts=integrations)
        self.assertIsNone(rc.get_proxies())

        integrations = { "integrations": { } }
        rc = RCObjectType(integrations, None)
        self.assertIsNone(rc.get_proxies())

        integrations = { "integrations": { "https_proxy": "abc" } }
        rc = RCObjectType(function_opts=None, opts=integrations)
        proxies = rc.get_proxies()
        self.assertEqual("abc", proxies['https'])
        self.assertIsNone(proxies['http'])

        integrations = { "integrations": { "https_proxy": "abc", 'http_proxy': 'def' } }
        rc = RCObjectType(integrations)
        proxies = rc.get_proxies()
        self.assertEqual("abc", proxies['https'])
        self.assertEqual("def", proxies['http'])

        os.environ["HTTP_PROXY"] = "https://mock.example.com:3128"
        integrations = { "integrations": { "https_proxy": "abc", 'http_proxy': 'def' } }
        rc = RCObjectType(integrations)
        proxies = rc.get_proxies()
        os.environ["HTTP_PROXY"] = ""
        self.assertEqual(None, proxies)

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    def test_timeout_overrides(self, RCObjectType):
        # test default timeout
        integrations = { "integrations": { } }
        rc = RCObjectType(integrations, None)
        self.assertEqual(rc.get_timeout(), 30)

        # test global setting
        integrations_timeout = { "integrations": { "timeout": "35" } }
        rc = RequestsCommon(integrations_timeout, None)
        self.assertEqual(rc.get_timeout(), 35)

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    def test_timeout_failure(self, RCObjectType):
        # test timeout
        integrations_twenty = { "integrations": { "timeout": "1" } }
        rc = RCObjectType(integrations_twenty, None)
        url = "/".join((TestFunctionRequests.URL_TEST_HTTP_VERBS, "delay", "10"))

        with self.assertRaises(IntegrationError):
            rc.execute_call_v2("get", url)

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    def test_timeout_success(self, RCObjectType):
        integrations_fourty = { "integrations": { "timeout": "5" } }
        rc = RCObjectType(integrations_fourty, None)
        url = "/".join((TestFunctionRequests.URL_TEST_HTTP_VERBS, "delay", "1"))
        resp = rc.execute_call_v2("get", url)
        assert resp.json() == {"delay": "1"}

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    def test_timeout_section_value(self, RCObjectType):
        # test section override of a global setting
        integrations_fourty = { "integrations": { "timeout": 40 } }
        integration_section = { "timeout": 50 }
        rc = RCObjectType(integrations_fourty, integration_section)
        self.assertEqual(rc.get_timeout(), 50)

    @unittest.skip(reason="https://api.ipify.org/ is currently unavailable")
    def test_resp_types(self):
        IPIFY = TestFunctionRequests.URL_TEST_DATA_RESULTS

        rc = RequestsCommon(None, None)

        # J S O N
        json_result = rc.execute_call("get", "{}?format=json".format(IPIFY), None)

        try:
            json.dumps(json_result)
        except (TypeError, OverflowError):
            self.fail("result is not json")

        # T E X T
        text_result = rc.execute_call("get", "{}?format=text".format(IPIFY), None, resp_type='text')
        self.assertIsNotNone(text_result)

        # B Y T E S
        bytes_result = rc.execute_call("get", "{}?format=text".format(IPIFY), None, resp_type='bytes')
        self.assertIsNotNone(bytes_result)
        self.assertTrue(isinstance(bytes_result, bytes))

    @unittest.skip(reason="https://api.ipify.org/ is currently unavailable")
    def test_v2_resp_type(self):
        IPIFY = TestFunctionRequests.URL_TEST_DATA_RESULTS

        rc = RequestsCommon(None, None)

        # R E S P O N S E  Object
        response = rc.execute_call_v2("get", "{}?format=json".format(IPIFY))

        self.assertTrue(isinstance(response, requests.models.Response))

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    @pytest.mark.livetest
    def test_verbs(self, RCObjectType):
        URL = TestFunctionRequests.URL_TEST_HTTP_VERBS

        headers = {
            "Content-type": "application/json; charset=UTF-8"
        }

        payload = {
            'title': 'foo',
            'body': 'bar',
            'userId': 1
        }

        rc = RCObjectType(None, None)


        # P O S T
        # test json argument without headers
        resp = rc.execute_call("post", "/".join((URL, "post")), payload, log=TestFunctionRequests.LOG)
        print (resp)
        self.assertEqual(resp['json'].get("body"), "bar")

        # test json argument with headers
        resp = rc.execute_call("post", "/".join((URL, "post")), payload, headers=headers, log=TestFunctionRequests.LOG)
        print (resp)
        self.assertEqual(resp['json'].get("body"), "bar")

        # test data argument
        headers_data = {
            "Content-type": "application/x-www-form-urlencoded"
        }
        resp = rc.execute_call("post", "/".join((URL, "post")), payload, headers=headers_data, log=TestFunctionRequests.LOG)
        print (resp)
        self.assertEqual(resp['json'].get("body"), "bar")


        # G E T
        resp = rc.execute_call("get", "/".join((URL, "get")), payload, log=TestFunctionRequests.LOG)
        self.assertTrue(resp['args'].get("userId"))
        self.assertEqual(resp['args'].get("userId"), '1')


        # P U T
        resp = rc.execute_call("put", "/".join((URL, "put")), payload, headers=headers, log=TestFunctionRequests.LOG)
        TestFunctionRequests.LOG.info(resp)
        self.assertTrue(resp['args'].get("title"))
        self.assertEqual(resp['args'].get("title"), 'foo')


        # P A T C H
        patch = {
            'title': 'patch'
        }

        resp = rc.execute_call("patch", "/".join((URL, "patch")), patch, headers=headers, log=TestFunctionRequests.LOG)
        print ("resp {}".format(resp))
        self.assertTrue(resp['args'].get("title"))
        self.assertEqual(resp['args'].get("title"), 'patch')

        # D E L E T E
        DEL_URL = "/".join((URL, "delete"))
        resp = rc.execute_call("delete", DEL_URL, None, log=TestFunctionRequests.LOG)
        self.assertEqual(resp.get("url"), DEL_URL)

        # bad verb
        with self.assertRaises(IntegrationError):
            resp = rc.execute_call("bad", URL, None, log=TestFunctionRequests.LOG)

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    @pytest.mark.livetest
    def test_verbs_v2(self, RCObjectType):
        URL = TestFunctionRequests.URL_TEST_HTTP_VERBS

        headers = {
            "Content-type": "application/json; charset=UTF-8"
        }

        payload = {
            'title': 'foo',
            'body': 'bar',
            'userId': 1
        }

        rc = RCObjectType(None, None)

        # P O S T
        # test json argument without headers
        resp = rc.execute("post", "/".join((URL, "post")), json=payload)
        print (resp.json())
        self.assertEqual(resp.json()["json"].get("body"), "bar")

        # test json argument with headers
        resp = rc.execute_call_v2("post", "/".join((URL, "post")), json=payload, headers=headers)
        print (resp.json())
        self.assertEqual(resp.json()['json'].get("body"), "bar")

        # test data argument
        headers_data = {
            "Content-type": "application/x-www-form-urlencoded"
        }
        resp = rc.execute("post", "/".join((URL, "post")), data=payload, headers=headers_data)
        print (resp.json())
        self.assertEqual(resp.json()['json'].get("body"), "bar")

        # G E T
        resp = rc.execute_call_v2("get", "/".join((URL, "get")), params=payload)
        self.assertTrue(resp.json()['args'].get("userId"))
        self.assertEqual(resp.json()['args'].get("userId"), '1')

        # P U T
        # With params
        resp = rc.execute("put", "/".join((URL, "put")), params=payload, headers=headers)
        TestFunctionRequests.LOG.info(resp)
        self.assertTrue(resp.json()['args'].get("title"))
        self.assertEqual(resp.json()['args'].get("title"), 'foo')

        # With json body
        resp = rc.execute_call_v2("put", "/".join((URL, "put")), json=payload, headers=headers)
        TestFunctionRequests.LOG.info(resp)
        self.assertTrue(resp.json()['json'].get("title"))
        self.assertEqual(resp.json()['json'].get("title"), 'foo')

        # P A T C H
        patch = {
            'title': 'patch'
        }
        # With params
        resp = rc.execute_call_v2("patch", "/".join((URL, "patch")), params=patch, headers=headers)
        print ("resp {}".format(resp.json()))
        self.assertTrue(resp.json()['args'].get("title"))
        self.assertEqual(resp.json()['args'].get("title"), 'patch')

        # With json body
        resp = rc.execute_call_v2("patch", "/".join((URL, "patch")), json=patch, headers=headers)
        print ("resp {}".format(resp.json()))
        self.assertTrue(resp.json()['json'].get("title"))
        self.assertEqual(resp.json()['json'].get("title"), 'patch')

        # D E L E T E
        DEL_URL = "/".join((URL, "delete"))
        resp = rc.execute_call_v2("delete", DEL_URL)
        self.assertEqual(resp.json().get("url"), DEL_URL)

        # bad verb
        with self.assertRaises(IntegrationError):
            resp = rc.execute_call_v2("bad", URL)

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    @pytest.mark.livetest
    def test_statuscode(self, RCObjectType):
        URL = TestFunctionRequests.URL_TEST_HTTP_STATUS_CODES

        rc = RCObjectType(None, None)

        resp = rc.execute_call("get", "/".join((URL, "200")), None, resp_type='text')

        with self.assertRaises(IntegrationError):
            resp = rc.execute_call("get", "/".join((URL, "300")), None, resp_type='text')

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    @pytest.mark.livetest
    def test_statuscode_v2(self, RCObjectType):
        URL = TestFunctionRequests.URL_TEST_HTTP_STATUS_CODES

        rc = RCObjectType(None, None)

        resp = rc.execute_call_v2("get", "/".join((URL, "200")))

        with self.assertRaises(IntegrationError):
            resp = rc.execute_call_v2("get", "/".join((URL, "400")))

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    @pytest.mark.livetest
    def test_statuscode_callback(self, RCObjectType):
        URL = "/".join((TestFunctionRequests.URL_TEST_HTTP_STATUS_CODES, "300"))

        def callback(resp):
            if resp.status_code != 300:
                raise ValueError(resp.status_code)

        rc = RCObjectType(None, None)

        resp = rc.execute_call("get", URL, None, resp_type='text', callback=callback)

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    @pytest.mark.livetest
    def test_statuscode_callback_v2(self, RCObjectType):
        URL = "/".join((TestFunctionRequests.URL_TEST_HTTP_STATUS_CODES, "300"))

        def callback(resp):
            if resp.status_code != 300:
                raise ValueError(resp.status_code)

        rc = RCObjectType(None, None)

        resp = rc.execute_call_v2("get", URL, callback=callback)

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    @pytest.mark.livetest
    def test_timeout(self, RCObjectType):
        URL = "/".join((TestFunctionRequests.URL_TEST_HTTP_STATUS_CODES, "200?sleep=30000"))

        rc = RCObjectType(None, None)

        with self.assertRaises(IntegrationError):
            resp = rc.execute_call("get", URL, None, resp_type='text', timeout=2)

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    @pytest.mark.livetest
    def test_timeout_v2(self, RCObjectType):
        URL = "/".join((TestFunctionRequests.URL_TEST_HTTP_STATUS_CODES, "200?sleep=30000"))

        rc = RCObjectType(None, None)

        with self.assertRaises(IntegrationError):
            resp = rc.execute_call_v2("get", URL, timeout=2)

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    @pytest.mark.livetest
    def test_basicauth(self, RCObjectType):
        URL = "/".join((TestFunctionRequests.URL_TEST_HTTP_VERBS, "basic-auth"))
        basicauth = ("postman", "password")

        rc = RCObjectType(None, None)

        resp = rc.execute_call("get", URL, None, basicauth=basicauth)
        self.assertTrue(resp.get("authenticated"))

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    @pytest.mark.livetest
    def test_basicauth_v2(self, RCObjectType):
        URL = "/".join((TestFunctionRequests.URL_TEST_HTTP_VERBS, "basic-auth"))
        basicauth = ("postman", "password")

        rc = RCObjectType(None, None)

        resp = rc.execute_call_v2("get", URL, auth=basicauth)
        self.assertTrue(resp.json().get("authenticated"))

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    @pytest.mark.livetest
    def test_proxy_override(self, RCObjectType):
        rc = RCObjectType(None, None)
        proxies = rc.get_proxies()
        self.assertIsNone(proxies)

        # test only integration proxies
        integrations_xyz =  {
            "integrations": {
                'http_proxy': "http://xyz.com",
                'https_proxy': "https://xyz.com"
            }
        }

        function_proxy_none =  {
            'http_proxy': None,
            'https_proxy': None
        }

        rc = RCObjectType(integrations_xyz)
        proxies = rc.get_proxies()
        self.assertEqual(proxies['http'], "http://xyz.com")
        self.assertEqual(proxies['https'], "https://xyz.com")

        rc = RCObjectType(integrations_xyz, function_proxy_none)
        proxies = rc.get_proxies()
        self.assertEqual(proxies['http'], "http://xyz.com")
        self.assertEqual(proxies['https'], "https://xyz.com")

        # test only function proxies
        integrations_none =  {
            "integrations": {
                'http_proxy': None,
                'https_proxy': None
            }
        }

        function_proxy_abc =  {
            'http_proxy': "http://abc.com",
            'https_proxy': "https://abc.com"
        }

        rc = RCObjectType(function_opts=function_proxy_abc)
        proxies = rc.get_proxies()
        self.assertEqual(proxies['http'], "http://abc.com")
        self.assertEqual(proxies['https'], "https://abc.com")

        rc = RCObjectType(integrations_none, function_proxy_abc)
        proxies = rc.get_proxies()
        self.assertEqual(proxies['http'], "http://abc.com")
        self.assertEqual(proxies['https'], "https://abc.com")


        # test integration and function proxies (override)
        rc = RCObjectType(integrations_xyz, function_proxy_abc)
        proxies = rc.get_proxies()
        self.assertEqual(proxies['http'], "http://abc.com")
        self.assertEqual(proxies['https'], "https://abc.com")

    @unittest.skip(reason="may be over the limit")
    def test_proxy(self):
        rc = RequestsCommon()

        proxy_url = self.URL_TEST_PROXY
        proxy_result = rc.execute_call("get", proxy_url, None)

        proxies = {
            'http': proxy_result['curl'] if proxy_result['protocol'] == 'http' else None,
            'https': proxy_result['curl'] if proxy_result['protocol'] == 'https' else None
        }

        URL = "?".join((TestFunctionRequests.URL_TEST_DATA_RESULTS, "format=json"))

        # J S O N
        json_result = rc.execute_call("get", URL, None, proxies=proxies)

        self.assertTrue(json_result.get("ip"))

        integrations =  { "integrations": {
            'http_proxy': proxy_result['curl'] if proxy_result['protocol'] == 'http' else None,
            'https_proxy': proxy_result['curl'] if proxy_result['protocol'] == 'https' else None
        }
        }

        rc = RequestsCommon(opts=integrations)
        json_result = rc.execute_call("get", URL, None)
        self.assertTrue(json_result.get("ip"))

    @unittest.skip(reason="may be over the limit")
    def test_proxy_v2(self):
        rc = RequestsCommon()

        proxy_url = TestFunctionRequests.URL_TEST_PROXY
        proxy_result = rc.execute_call_v2("get", proxy_url)
        proxy_result_json = proxy_result.json()

        proxies = {
            'http': proxy_result_json['curl'] if proxy_result_json['protocol'] == 'http' else None,
            'https': proxy_result_json['curl'] if proxy_result_json['protocol'] == 'https' else None
        }

        URL = "?".join((TestFunctionRequests.URL_TEST_DATA_RESULTS, "format=json"))

        # J S O N
        response = rc.execute_call_v2("get", URL, proxies=proxies)
        json_result = response.json()

        self.assertTrue(json_result.get("ip"))

        integrations =  { "integrations": {
            'http_proxy': proxy_result_json['curl'] if proxy_result_json['protocol'] == 'http' else None,
            'https_proxy': proxy_result_json['curl'] if proxy_result_json['protocol'] == 'https' else None
        }
        }

        rc = RequestsCommon(opts=integrations)
        response = rc.execute_call_v2("get", URL)
        json_result = response.json()
        self.assertTrue(json_result.get("ip"))

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    @pytest.mark.livetest
    def test_headers(self, RCObjectType):
        # G E T with headers
        headers = {
            "Content-type": "application/json; charset=UTF-8",
            "my-sample-header": "my header"
        }
        URL = "/".join((TestFunctionRequests.URL_TEST_HTTP_VERBS, "headers"))

        rc = RCObjectType()

        json_result = rc.execute_call("get", URL, None, headers=headers)
        self.assertEqual(json_result['headers'].get("my-sample-header"), "my header")

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    @pytest.mark.livetest
    def test_headers_v2(self, RCObjectType):
        # G E T with headers
        headers = {
            "Content-type": "application/json; charset=UTF-8",
            "my-sample-header": "my header"
        }
        URL = "/".join((TestFunctionRequests.URL_TEST_HTTP_VERBS, "headers"))

        rc = RCObjectType()

        json_result = rc.execute_call_v2("get", URL, headers=headers)
        self.assertEqual(json_result.json()['headers'].get("my-sample-header"), "my header")


    @parameterized.expand([
        [None, True],
        ["application/json", True],
        ["application/json; charset=UTF-8", True],
        ["charset=UTF-8; application/json", True],
        ["application/x-www-form-urlencoded", False]
    ])
    def test_is_payload_in_json(self, content_type, result):

        payload_in_json = is_payload_in_json(content_type)

        self.assertEqual(payload_in_json, result)


    @parameterized.expand([
        [None, None],
        [{"Content-Type": "mock_type"}, "mock_type"],
        [{"my-sample-header": "my header"}, None]
    ])
    def test_get_case_insensitive_key_value(self, dictionary, result):

        value = get_case_insensitive_key_value(dictionary, "content-type")

        self.assertEqual(value, result)

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    def test_timeout_override(self, RCObjectType):
        rc = RCObjectType(None, None)
        timeout = rc.get_timeout()
        self.assertEqual(timeout, 30) # default in get_timeout()

        # test only integration proxies
        integrations_60 =  {
            "integrations": {
                'timeout': "60"
            }
        }

        function_timeout_none =  {
        }

        rc = RCObjectType(integrations_60)
        timeout = rc.get_timeout()
        self.assertEqual(timeout, 60)

        rc = RCObjectType(integrations_60, function_timeout_none)
        timeout = rc.get_timeout()
        self.assertEqual(timeout, 60)

        # test only function timeout
        integrations_none =  {
            "integrations": {
            }
        }

        function_timeout_90 =  {
            'timeout': "90"
        }

        rc = RCObjectType(function_opts=function_timeout_90)
        timeout = rc.get_timeout()
        self.assertEqual(timeout, 90)

        rc = RCObjectType(integrations_none, function_timeout_90)
        timeout = rc.get_timeout()
        self.assertEqual(timeout, 90)

        # test integration and function proxies (override)
        rc = RCObjectType(integrations_60, function_timeout_90)
        timeout = rc.get_timeout()
        self.assertEqual(timeout, 90)

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    def test_clientauth(self, RCObjectType):
        rc = RCObjectType()
        cert = rc.get_clientauth()
        self.assertIsNone(cert)


        mock_fn_section = {
            "url": "fake_url.com",
            "client_auth_cert": "cert.pem",
            "client_auth_key": "private.pem"
        }
        rc = RCObjectType(opts=None, function_opts=mock_fn_section)
        cert = rc.get_clientauth()
        self.assertEqual(cert, ("cert.pem", "private.pem"))

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    def test_client_auth(self, RCObjectType):
        rc = RCObjectType()
        cert = rc.get_client_auth()
        self.assertIsNone(cert)


        mock_fn_section = {
            "url": "fake_url.com",
            "client_auth_cert": "cert.pem",
            "client_auth_key": "private.pem"
        }
        rc = RCObjectType(opts=None, function_opts=mock_fn_section)
        cert = rc.get_client_auth()
        self.assertEqual(cert, ("cert.pem", "private.pem"))


        mock_fn_section = {
            "url": "fake_url.com",
            "client_auth_cert": "cert.pem"
        }
        rc = RCObjectType(opts=None, function_opts=mock_fn_section)
        cert = rc.get_client_auth()
        self.assertIsNone(cert)


        mock_fn_section = {
            "url": "fake_url.com",
            "client_auth_cert": mock_paths.MOCK_CLIENT_CERT_FILE,
            "client_auth_key": mock_paths.MOCK_CLIENT_KEY_FILE
        }
        rc = RCObjectType(opts=None, function_opts=mock_fn_section)
        # make sure can still make call to execute with no issues
        resp = rc.execute("get", self.URL_TEST_HTTP_VERBS)
        self.assertEqual(resp.status_code, 200)


        rc = RCObjectType()
        # make sure can call execute with clientauth optional parameter
        # but doesn't set the cert value for the whole object
        resp = rc.execute("get", self.URL_TEST_HTTP_VERBS, clientauth=(mock_paths.MOCK_CLIENT_CERT_FILE, mock_paths.MOCK_CLIENT_KEY_FILE))
        self.assertEqual(resp.status_code, 200)
        cert = rc.get_client_auth()
        self.assertIsNone(cert)

    @parameterized.expand(REQUESTS_COMMON_CLASSES)
    def test_get_no_verify(self, RCObjectType):
        rc = RCObjectType()
        verify = rc.get_verify()
        self.assertIsNone(verify)

    @parameterized.expand([
        [{"verify": "false"}, False, RequestsCommon],
        [{"verify": False}, False, RequestsCommon],
        [{"verify": "true"}, True, RequestsCommon],
        [{"verify": True}, True, RequestsCommon],
        [{"verify": "path_to_CA_bundle"}, "path_to_CA_bundle", RequestsCommon],
        [{"verify": "false"}, False, RequestsCommonWithoutSession],
        [{"verify": False}, False, RequestsCommonWithoutSession],
        [{"verify": "true"}, True, RequestsCommonWithoutSession],
        [{"verify": True}, True, RequestsCommonWithoutSession],
        [{"verify": "path_to_CA_bundle"}, "path_to_CA_bundle", RequestsCommonWithoutSession],
    ])
    def test_get_verify_in_app_section(self, mock_fn_section, expected_verify, RCObjectType):
        rc = RCObjectType(opts=None, function_opts=mock_fn_section)
        verify = rc.get_verify()
        self.assertEqual(verify, expected_verify)

    @parameterized.expand([
        [{"integrations":{"verify": "path_to_CA_bundle"}}, {}, "path_to_CA_bundle", RequestsCommon],
        [{"integrations":{"verify": "true"}}, {}, True, RequestsCommon],
        [{"integrations":{"verify": "false"}}, {}, False, RequestsCommon],
        [{"integrations":{"verify": False}}, {}, False, RequestsCommon],
        [{"integrations":{"verify": "false"}}, {"verify": "path_to_bundle"}, "path_to_bundle", RequestsCommon],
        [{"integrations":{"verify": "path_to_CA_bundle"}}, {}, "path_to_CA_bundle", RequestsCommonWithoutSession],
        [{"integrations":{"verify": "true"}}, {}, True, RequestsCommonWithoutSession],
        [{"integrations":{"verify": "false"}}, {}, False, RequestsCommonWithoutSession],
        [{"integrations":{"verify": False}}, {}, False, RequestsCommonWithoutSession],
        [{"integrations":{"verify": "false"}}, {"verify": "path_to_bundle"}, "path_to_bundle", RequestsCommonWithoutSession]
    ])
    def test_get_verify_in_integrations_section(self, opts, function_opts, expected_verify, RCObjectType):
        rc = RCObjectType(opts=opts, function_opts=function_opts)
        verify = rc.get_verify()
        self.assertEqual(verify, expected_verify)

    @parameterized.expand([
        # mostly here for backward compatibility with older instances of apps using of rc.execute()
        # with a value of `verify` passed in
        # and make sure that the value passed in directly is used over any configs
        ["https://example.com", {}, {"verify": "False"}, True, True, RequestsCommon],
        ["https://example.com", {"integrations": {"verify": True}}, {}, False, False, RequestsCommon],
        ["https://example.com", {}, {}, None, True, RequestsCommon],
        ["https://example.com", {}, {}, "path_to_bundle", "path_to_bundle", RequestsCommon],

        # and make sure that if no value is passed, then the configs are used
        ["https://example.com", {}, {"verify": "False"}, None, False, RequestsCommon],
        ["https://example.com", {"integrations": {"verify": True}}, {}, None, True, RequestsCommon],
        ["https://example.com", {"integrations": {"verify": True}}, {"verify": "False"}, None, False, RequestsCommon],

        # Repeat all tests with RequestsCommonWithoutSession
        ["https://example.com", {}, {"verify": "False"}, True, True, RequestsCommonWithoutSession],
        ["https://example.com", {"integrations": {"verify": True}}, {}, False, False, RequestsCommonWithoutSession],
        ["https://example.com", {}, {}, None, True, RequestsCommonWithoutSession],
        ["https://example.com", {}, {}, "path_to_bundle", "path_to_bundle", RequestsCommonWithoutSession],
        ["https://example.com", {}, {"verify": "False"}, None, False, RequestsCommonWithoutSession],
        ["https://example.com", {"integrations": {"verify": True}}, {}, None, True, RequestsCommonWithoutSession],
        ["https://example.com", {"integrations": {"verify": True}}, {"verify": "False"}, None, False, RequestsCommonWithoutSession]
    ])
    def test_execute_request_with_verify(self, url, opts, function_opts, verify, expected_verify, RCObjectType):
        # this check is to ensure that calling rc.execute() is properly grabbing
        # the value that is given directly to it.
        # The main reason to run this test, is to ensure backward compatiblity
        # with apps already using rc.execute() with the `verify` parameter set.
        # in those cases, we want to make sure that we don't override anything with
        # in the app.config, but instead continue to use it as the developer of that
        # app expected it to work


        # register a request mocker to intercept and monitor the requests
        with requests_mock.Mocker() as m:
            m.get(url)
            rc = RCObjectType(opts=opts, function_opts=function_opts)
            rc.execute("GET", url, verify=verify)

            # assert that the used value for verify was what we expected
            assert m.request_history[0].verify == expected_verify


    def test_sessions_cookies(self):
        """
        This test proves that the "With Session" version of the RC object
        properly uses the session by retaining cookies from the first request
        NOTE: this is a live test
        """

        # Test with session, where cookies should persist
        rc = RequestsCommon()
        rc.execute("GET", "{0}/cookies/set?foo=bar&jon=snow".format(self.URL_TEST_HTTP_VERBS))
        resp = rc.execute("GET", "{0}/cookies".format(self.URL_TEST_HTTP_VERBS))

        assert "cookies" in resp.json()
        assert resp.json()["cookies"] == {"foo": "bar", "jon": "snow"}

        # Test again with normal RC object, where cookies won't persist
        rc = RequestsCommonWithoutSession()
        rc.execute("GET", "{0}/cookies/set?foo=bar&jon=snow".format(self.URL_TEST_HTTP_VERBS))
        resp = rc.execute("GET", "{0}/cookies".format(self.URL_TEST_HTTP_VERBS))

        assert "cookies" in resp.json()
        assert resp.json()["cookies"] == {}

    @unittest.skip(reason="generally true, but not regular enough so closing for now")
    def test_sessions_faster_than_regular(self):
        """
        This test proves that the session object is much more efficient when
        hitting the same endpoint.
        NOTE: this is a live test
        """

        # N needs to be large enough to see a difference (minimum 25 or so)
        # that said, the difference is pretty start even with small iteration numbers
        N = 25

        rc = RequestsCommon()
        start = time.time()
        for i in range(N):
            rc.execute("GET", "{0}/basic-auth".format(self.URL_TEST_HTTP_VERBS),
                        headers={"Authorization": "Basic cG9zdG1hbjpwYXNzd29yZA=="})
        end = time.time()
        session_time = end - start

        rc = RequestsCommonWithoutSession()
        start = time.time()
        for i in range(N):
            rc.execute("GET", "{0}/basic-auth".format(self.URL_TEST_HTTP_VERBS),
                        headers={"Authorization": "Basic cG9zdG1hbjpwYXNzd29yZA=="})
        end = time.time()
        standard_time = end - start

        assert session_time < standard_time



"""
Enables conditions to be used with different UI elements in the Tab definition.
In a way like:

class TestTab(Tab):
    ...
    SHOW_IF = [
        Field('qradar_id').conditions.has_value(),
        Field('desription').conditions.contains('ATP')
    ]
"""
class Conditions(object):
    def __init__(self, api_name):
        self.api_name = api_name


class FieldConditions(Conditions):
    def has_value(self):
        return {
            "field": "incident.{}".format(self.api_name),
            "condition": "has_a_value"
        }

    def equals(self, value):
        return {
            "field": "incident.{}".format(self.api_name),
            "condition": "equals",
            "value": value
        }

    def contains(self, value):
        return {
            "field": "incident.{}".format(self.api_name),
            "condition": "contains",
            "value": value
        }

    def doesnt_contain(self, value):
        return {
            "field": "incident.{}".format(self.api_name),
            "condition": "not_contains",
            "value": value
        }

    def doesnt_equal(self, value):
        return {
            "field": "incident.{}".format(self.api_name),
            "condition": "not_equals",
            "value": value
        }

    def doesnt_have_value(self, value):
        return {
            "field": "incident.{}".format(self.api_name),
            "condition": "not_has_a_value"
        }



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2020. All Rights Reserved.
# pragma pylint: disable=unused-argument, no-self-use

import json
from .function_metrics import FunctionMetrics

PAYLOAD_VERSION = "1.0"

class ResultPayload:
    """ Class to create a standard payload for functions. The resulting payload follows the following format:
        1.0
        { "version": "1.0"       -- used to track different versions of the payload
          "success": True|False
          "reason": str          -- a string to explain if success=False
          "content": json        -- the result of the function call
          "raw": str             -- a string representation of content. This is sometimes needed when the result of one function is
                                    piped into the next
          "inputs": json         -- a copy of the input parameters, useful for post-processor script use
          "metrics": json        -- a set of information to capture specifics metrics about the function's runtime environment
        }
    """
    def __init__(self, pkgname, version=PAYLOAD_VERSION, **kwargs):
        """
        build initial payload data structure and the start the timers for metrics collection
        :param pkgname: package name to capture stats on which package is being used
        :param kwargs: input parameters for this function
        """
        self.fm = FunctionMetrics(pkgname)
        self.payload = {
            "version": version,
            "success": None,
            "reason":  None,
            "content": None,
            "raw": None,
            "inputs":  kwargs,
            "metrics": None
        }

    def done(self, success, content, reason=None):
        """
         complete the function payload
        :param success: True|False
        :param content: json result to pass back
        :param reason: comment fields when success=False
        :return: completed payload in json
        """
        self.payload['success'] = success
        self.payload['reason'] = reason
        self.payload['content'] = content
        self.payload['metrics'] = self.fm.finish()

        if float(self.payload.get("version", 2.0)) < 2.0:
            try:
                self.payload['raw'] = json.dumps(content)
            except:
                pass

        return self.payload



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2020. All Rights Reserved.
# pragma pylint: disable=unused-argument, no-self-use
from .conditions import Conditions, FieldConditions


class UIElementBase(object):
    ELEMENT_TYPE = None

    def __init__(self, api_name):
        assert self.ELEMENT_TYPE, "Element type for subclass needs to be defined"
        self.api_name = api_name
        self.conditions = Conditions(self.api_name)

    def as_dto(self):
        """
        Returns a JSON serializable dictionary with DTO representing the field in the tab.
        """
        raise NotImplementedError("DTO represenation for {} is not implemented".format(self.__class__))

    def exists_in(self, fields):
        """
        Given a list of fields of a tab, find if current one is one of them.
        """
        return next((field for field in fields
                    if field["element"] == self.ELEMENT_TYPE and field["content"] == self.api_name), None) is not None


class Field(UIElementBase):
    ELEMENT_TYPE = "field"

    def __init__(self, api_name):
        super(Field, self).__init__(api_name)
        self.conditions = FieldConditions(api_name)

    def as_dto(self):
        return {
            "content": self.api_name,
            "element": self.ELEMENT_TYPE,
            "field_type": "incident"
        }


class Datatable(UIElementBase):
    ELEMENT_TYPE = "datatable"

    def as_dto(self):
        return {
            "element": self.ELEMENT_TYPE,
            "content": self.api_name
        }

class View(UIElementBase):
    ELEMENT_TYPE = "view"

    def as_dto(self):
        return {
            "element": self.ELEMENT_TYPE,
            "content": self.api_name
        }



import pytest
from mock import patch
from resilient_lib.ui import Tab, Datatable, Field
from resilient_lib.ui import create_tab


class MockTab(Tab):
    NAME = "test"
    UUID = "test"
    SECTION = "test"
    CONTAINS = [
        Datatable('test'),
        Field('test')
    ]

    SHOW_IF = [Field('test').conditions.has_value()]

class TestSubmittedData(object):
    """
    Tests that payload submitted to the server contains proper tab information.
    """

    @pytest.mark.skip(reason="Needs better mocking")
    @patch("resilient.get_client")
    def test_create_tab(self, get_client):
        # this mocks the requests made to /types and /layout?type_id=xxx
        get_client.return_value.get.side_effect = [
            {
                "organization": {
                    "type_id": 42
                }
            },
            [{
                "id": 42,
                "name": "incident",
                "content": []
            }]
        ]

        create_tab(MockTab, {})

        # assert that PUT was called and correct payload present
        get_client.return_value.put.assert_called_once()
        call_args = get_client.return_value.put.call_args
        payload = call_args.kwargs.get("payload")
        assert MockTab.exists_in(payload.get('content'))
        for field in MockTab.CONTAINS:
            assert field.exists_in(MockTab.get_from_tabs(payload.get('content')).get("fields"))

    @pytest.mark.skip(reason="Needs better mocking")
    @patch("resilient.get_client")
    def test_create_tab_with_locked_ui(self, get_client):

        mock_opts = {
            "resilient": {
                "ui_lock": "True"
            },
            MockTab.SECTION: {}
        }

        # this mocks the requests made to /types and /layout?type_id=xxx
        get_client.return_value.get.side_effect = [
            {
                "organization": {
                    "type_id": 42
                }
            },
            [{
                "id": 42,
                "name": "incident",
                "content": []
            }]
        ]

        create_tab(MockTab, mock_opts)

        # assert that PUT was called and correct payload present
        assert get_client.return_value.put.call_count == 0

    @pytest.mark.skip(reason="Needs better mocking")
    @patch("resilient.get_client")
    def test_update_tab_disabled(self, get_client):

        # this mocks the requests made to /types and /layout?type_id=xxx
        get_client.return_value.get.side_effect = [
            {
                "organization": {
                    "type_id": 42
                }
            },
            [{
                "id": 42,
                "name": "incident",
                "content": [{
                    "predefined_uuid": MockTab.UUID,
                    "fields": [
                        Field("test").as_dto()
                    ]
                }]
            }]
        ]

        create_tab(MockTab, {})

        # assert that PUT was called and correct payload present
        get_client.return_value.put.call_count == 0

    @pytest.mark.skip(reason="Needs better mocking")
    @patch("resilient.get_client")
    def test_update_tab_enabled(self, get_client):

        # this mocks the requests made to /types and /layout?type_id=xxx
        get_client.return_value.get.side_effect = [
            {
                "organization": {
                    "type_id": 42
                }
            },
            [{
                "id": 42,
                "name": "incident",
                "content": [{
                    "predefined_uuid": MockTab.UUID,
                    "fields": [
                        Field("test").as_dto()
                    ]
                }]
            }]
        ]

        create_tab(MockTab, {}, update_existing=True)

        get_client.return_value.put.assert_called_once()
        call_args = get_client.return_value.put.call_args
        payload = call_args.kwargs.get("payload")
        assert MockTab.exists_in(payload.get('content'))
        for field in MockTab.CONTAINS:
            assert field.exists_in(MockTab.get_from_tabs(payload.get('content')).get("fields"))

    @pytest.mark.skip(reason="Needs better mocking")
    @patch("resilient.get_client")
    def test_conditions_sent(self, get_client):

        # this mocks the requests made to /types and /layout?type_id=xxx
        get_client.return_value.get.side_effect = [
            {
                "organization": {
                    "type_id": 42
                }
            },
            [{
                "id": 42,
                "name": "incident",
                "content": [{
                    "predefined_uuid": MockTab.UUID,
                    "fields": [
                        Field("test").as_dto()
                    ]
                }]
            }]
        ]

        create_tab(MockTab, {}, update_existing=True)

        get_client.return_value.put.assert_called_once()
        call_args = get_client.return_value.put.call_args
        payload = call_args.kwargs.get("payload")
        assert MockTab.exists_in(payload.get('content'))
        for field in MockTab.CONTAINS:
            assert field.exists_in(MockTab.get_from_tabs(
                payload.get('content')).get("fields"))
        assert MockTab.get_from_tabs(payload.get('content')).get('show_if') == MockTab.SHOW_IF



import unittest
from resilient_lib.components.html2markdown import MarkdownParser

class TestFunctionMetrics(unittest.TestCase):
    """ Tests for the attachment_hash function"""

    def test_no_html(self):

        parser = MarkdownParser()
        data =  """The General Preferences Pane allows you to tell me how you want me to behave. 
For example, do you want me to make sure there is a document open when I launch?
You can also tell me if I should constantly update the preview window as you type, or wait for you to hit command-R instead.
Maybe you prefer your editor window on the right? Or to see the word-count as you type. 
This is also the place to tell me if you are interested in pre-releases of me, or just want to stick to better-tested official releases."""

        converted = parser.convert(data)
        self.assertEqual(converted, data)

    def test_paragraph(self):
        data = """<div class="rte"><div>this is line 1</div><div>this is line 2</div><div>this is line 3</div></div>"""
        markdown = "this is line 1\nthis is line 2\nthis is line 3"

        parser = MarkdownParser()
        converted= parser.convert(data)
        self.assertEqual(converted, markdown)

    def test_blockquote(self):
        data = """<div class="rte"><blockquote>this is line 1
this is line 2
this is line 3</blockquote></div>"""
        markdown = "```this is line 1\nthis is line 2\nthis is line 3```"

        parser = MarkdownParser()
        converted= parser.convert(data)
        self.assertEqual(converted, markdown)

    def test_anchor(self):
        parser = MarkdownParser()
        data = """<div>
<a href="https://ibm/biz/resilientcommunity" target="_blank">anchor</a>
</div>"""
        data_monospace = """<div style="text-align: justify;">
<a href="https://ibm/biz/resilientcommunity" style="font-size: 10.0px;font-family: monospace;" target="_blank">anchor</a>
</div>"""
        markdown="[https://ibm/biz/resilientcommunity](anchor)"
        markdown_monospace="[https://ibm/biz/resilientcommunity]({{anchor}})"
        converted = parser.convert(data)
        self.assertEqual(converted, markdown)

        parser_monospace = MarkdownParser()
        converted_monospace = parser_monospace.convert(data_monospace)
        self.assertEqual(converted_monospace, markdown_monospace)

    def test_strong_emphasis_underline_strikeout(self):
        data = """<div><strong>bold</strong><span> </span><em>italic</em><span> </span><u>underline</u><span> </span><s>strikeout</s></div>"""
        markdown = "**bold** *italic* __underline__ ~~strikeout~~"

        parser = MarkdownParser()
        converted = parser.convert(data)
        self.assertEqual(converted, markdown)

    def test_underline_and_emphasis(self):
        data = "<div class='rte'><div><strong><u>underline and strong</u></strong></div></div>"
        markdown = "**__underline and strong__**"

        parser = MarkdownParser()
        converted = parser.convert(data)
        self.assertEqual(converted, markdown)

        data = "<div class='rte'><div style='font-family: monospace; color: rgb(0, 255, 0)'><strong><u>underline and strong</u></strong></div></div>"
        markdown = "{color:#00ff00}{{**__underline and strong__**}}{color}"

        parser = MarkdownParser()
        converted = parser.convert(data)
        self.assertEqual(converted, markdown)

    def test_color(self):
        data = """<div class='rte'><span style="color: rgb(255,0,0);">this is red</span></div>"""
        markdown = "{color:#ff0000}this is red{color}"

        parser = MarkdownParser()
        converted = parser.convert(data)
        self.assertEqual(converted, markdown)

        # same test but using str() which calls __str__()
        self.assertEqual(str(parser), markdown)

    def test_h(self):
        data = """<div><h1>this is a header</h1><strong>strong</strong></div>"""
        markdown = "h1. this is a header\n**strong**"

        parser = MarkdownParser()
        parser.convert(data)
        self.assertEqual(str(parser), markdown)

        data = """<div><h1>this is a header</h1><h2>h2</h2></div>"""
        markdown = "* this is a header\n** h2"

        parser = MarkdownParser(headers=['*', '**', '***'])
        parser.convert(data)
        self.assertEqual(str(parser), markdown)

    def test_unknown(self):
        data = """<div><x>this is a header</x><strong>strong</strong></div>"""
        markdown = "this is a header\n**strong**"

        parser = MarkdownParser()
        parser.convert(data)
        self.assertEqual(str(parser), markdown)

    def test_mismatch_tag(self):
        data = """<div><h1>this is a header</h1><strong>strong</strong></xx>"""

        parser = MarkdownParser()

        with self.assertRaises(ValueError):
            parser.convert(data)

    def test_missing_tag(self):
        data = """<div><h1>this is a header</h1><strong>strong</div>"""

        parser = MarkdownParser()

        with self.assertRaises(ValueError):
            parser.convert(data)


    def test_lists(self):
        data_ordered = """<div><ol><li>1</li><li>2</li><li>3</li></ol></div>"""
        markdown_ordered = """
    1. 1
    2. 2
    3. 3"""

        data_unordered = """<div><ul><li>1</li><li>2</li><li>3</li></ul></div>"""
        markdown_unordered = """
    * 1
    * 2
    * 3"""

        data_nested_ol_ol = """<div><ol><li>1</li><li>2</li><li>3</li><ol><li>31</li><li>32</li><li>33</li></ol><li>4</li></ol></div>"""
        markdown_ol_ol = """
    1. 1
    2. 2
    3. 3
        1. 31
        2. 32
        3. 33
    4. 4"""

        data_nested_ul_ol = """<div><ul><li>1</li><li>2</li><li>3</li><ol><li>31</li><li>32</li><li>33</li></ol><li>4</li></ul></div>"""
        markdown_ul_ol = """
    * 1
    * 2
    * 3
        1. 31
        2. 32
        3. 33
    * 4"""

        data_nested_ul_ul = """<div><ul><li>1</li><li>2</li><li>3</li><ul><li>31</li><li>32</li><li>33</li></ul><li>4</li></ul></div>"""
        markdown_ul_ul = """
    * 1
    * 2
    * 3
        * 31
        * 32
        * 33
    * 4"""

        data_nested_ol_ul = """<div><ol><li>1</li><li>2</li><li>3</li><ul><li>31</li><li>32</li><li>33</li></ul><li>4</li></ol></div>"""
        markdown_ol_ul = """
    1. 1
    2. 2
    3. 3
        * 31
        * 32
        * 33
    4. 4"""

        data_nested_ol_ul_ol = """<div class="rte"><ol><li>1</li><li>2</li><li>3</li></ol><ul><ul><li>a</li><li>b</li><li>c</li></ul></ul><ol><ol><ol><li>i</li><li>ii</li><li>iii</li></ol></ol></ol></div>"""
        markdown_ol_ul_ol = """
    1. 1
    2. 2
    3. 3
        * a
        * b
        * c
            1. i
            2. ii
            3. iii"""

        parser = MarkdownParser()
        converted_ordered = parser.convert(data_ordered)
        self.assertEqual(converted_ordered, markdown_ordered)

        parser = MarkdownParser()
        converted_unordered = parser.convert(data_unordered)
        self.assertEqual(converted_unordered, markdown_unordered)

        parser = MarkdownParser()
        converted_ol_ol = parser.convert(data_nested_ol_ol)
        self.assertEqual(converted_ol_ol, markdown_ol_ol)

        parser = MarkdownParser()
        converted_ul_ol = parser.convert(data_nested_ul_ol)
        self.assertEqual(converted_ul_ol, markdown_ul_ol)

        parser = MarkdownParser()
        converted_ul_ul = parser.convert(data_nested_ul_ul)
        self.assertEqual(converted_ul_ul, markdown_ul_ul)

        parser = MarkdownParser()
        converted_ol_ul = parser.convert(data_nested_ol_ul)
        self.assertEqual(converted_ol_ul, markdown_ol_ul)

        parser = MarkdownParser()
        converted_ol_ul_ol = parser.convert(data_nested_ol_ul_ol)
        self.assertEqual(converted_ol_ul_ol, markdown_ol_ul_ol)

        data_custom = """<div class="rte"><ul><li>1</li><li>2</li><li>3</li></ul><ul><ul><li>a</li><li>b</li><li>c</li></ul></ul><ul><ul><ul><li>i</li><li>ii</li><li>iii</li></ul></ul></ul></div>"""
        markdown_custom = """
    * 1
    * 2
    * 3
        + a
        + b
        + c
            - i
            - ii
            - iii"""

        parser = MarkdownParser(bullets=["*", "+", "-"])
        converted_custom = parser.convert(data_custom)
        self.assertEqual(converted_custom, markdown_custom)

        markdown_custom_single = """
* 1
* 2
* 3
* a
* b
* c
* i
* ii
* iii"""

        parser = MarkdownParser(bullets="*", indent=0)
        converted_custom = parser.convert(data_custom)
        self.assertEqual(converted_custom, markdown_custom_single)

    def test_end_list(self):
        data = u'<div class="rte"><div><strong>rich text</strong></div><div><br /></div><ol><li>111</li><li>2222</li><li>333</li><li>444</li></ol><div>normal</div></div>'
        data_result = "**rich text**\n\n\n\n\n    1. 111\n    2. 2222\n    3. 333\n    4. 444\n\nnormal"

        parser = MarkdownParser()
        converted_custom = parser.convert(data)
        self.assertEqual(converted_custom, data_result)

    def test_modify_symbols(self):
        data = "<div class='rte'><div><strong><u>underline and strong</u></strong></div></div>"
        markdown = "*_underline and strong_*"

        parser = MarkdownParser(bold="*", underline="_")
        converted = parser.convert(data)
        self.assertEqual(converted, markdown)

        data_unordered = """<div><ul><li>1</li><li>2</li><li>3</li></ul></div>"""
        markdown_unordered = """
  + 1
  + 2
  + 3"""

        parser = MarkdownParser(indent=2, bullets='+')
        converted_unordered = parser.convert(data_unordered)
        self.assertEqual(converted_unordered, markdown_unordered)

        data_ordered = """<div><ol><li>1</li><li>2</li><li>3</li></ol></div>"""
        markdown_ordered = """
    # 1
    # 2
    # 3"""

        parser = MarkdownParser(number='#')
        converted_ordered = parser.convert(data_ordered)
        self.assertEqual(converted_ordered, markdown_ordered)

    def test_monospace(self):
        data = """<div class="rte"><div><strong style="font-family: monospace;">monospace</strong><strong> and bold</strong></div><div><u>underline</u></div><div><strong><u>bold and underline</u></strong></div><div><br /></div></div>"""
        markdown_monospace = """**{{monospace}}**** and bold**\n__underline__\n**__bold and underline__**"""

        parser = MarkdownParser(number='#')
        converted_monospace = parser.convert(data)
        self.assertEqual(converted_monospace, markdown_monospace)

    def test_mixed(self):
        data = """<div class="rte"><div>this is a line with <strong>bold</strong> and <span style="color: rgb(255,0,0);">red</span></div></div>"""
        markdown_mixed="""this is a line with **bold** and {color:#ff0000}red{color}"""

        parser = MarkdownParser()
        converted_mixed = parser.convert(data)
        self.assertEqual(converted_mixed, markdown_mixed)

    def test_nested(self):
        data = """<div class="rte"><ol><li>111</li><li>222<ol><li>aaa</li><li>bbb</li></ol></li><li>333</li></ol></div>"""
        markdown = "\n    1. 111\n    2. 222\n        1. aaa\n        2. bbb\n    3. 333"

        parser = MarkdownParser(bullets=["*", "+", "-"])
        converted_mixed = parser.convert(data)
        self.assertEqual(converted_mixed, markdown)



# -*- coding: utf-8 -*-
# pragma pylint: disable=unused-argument, no-self-use
# (c) Copyright IBM Corp. 2010, 2022. All Rights Reserved.

import base64
import datetime
import functools
import logging
import traceback
from ast import literal_eval
from threading import Event

from cachetools import LRUCache, cached
from resilient import Patch, SimpleHTTPException
from resilient_lib import (IntegrationError, clean_html, get_file_attachment,
                           get_file_attachment_name)
from six import raise_from

LOG = logging.getLogger(__name__)

TYPES_URI = "/types"
INCIDENTS_URI = "/incidents"
ARTIFACTS_URI = "/artifacts"
ARTIFACT_FILE_URI = "/".join([ARTIFACTS_URI, "files"])

DEFAULT_CASES_QUERY_FILTER = "return_level=normal"

# P O L L E R   L O G I C
def poller(named_poller_interval, named_last_poller_time):
    """
    Decorator to wrap a function as the logic for a poller.

    **Example:**

    .. code-block:: python

        import logging
        from threading import Thread

        from resilient_circuits import AppFunctionComponent, is_this_a_selftest
        from resilient_lib import poller, get_last_poller_date

        PACKAGE_NAME = "fn_my_app"

        LOG = logging.getLogger(__name__)

        class MyPoller(AppFunctionComponent):

            def __init__(self, opts):
                super(PollerComponent, self).__init__(opts, PACKAGE_NAME)

                polling_interval = 5 # set to 5 seconds or could get from ``self.options``
                last_poller_time = get_last_poller_date(120) # look back 2 hours

                if is_this_a_selftest(self):
                    LOG.warn("Running selftest -- disabling poller")
                else:
                    poller_thread = Thread(target=self.run)
                    poller_thread.daemon = True
                    poller_thread.start()

            @poller('polling_interval', 'last_poller_time')
            def run(self, *args, **kwargs):

                # poll endpoint
                query_results = query_entities_since_last_poll(kwargs['last_poller_time'])

                # process any results to create, update, or close case in SOAR
                if query_results:
                    self.process_entity_list(query_results)


    :param named_poller_interval: name of instance variable containing the poller interval in seconds
    :type named_poller_interval: str
    :param named_last_poller_time: name of instance variable containing the lookback value in mseconds
    :type named_last_poller_time: datetime
    """
    def poller_wrapper(func):
        # decorator for running a function forever, passing the ms timestamp of
        #  when the last poller run to the function it's calling
        @functools.wraps(func)
        def wrapped(self, *args):

            last_poller_time = getattr(self, named_last_poller_time)
            exit_event = Event()

            while not exit_event.is_set():
                try:
                    LOG.info("%s polling start.", self.PACKAGE_NAME)
                    poller_start = datetime.datetime.now()
                    # function execution with the last poller time in ms
                    func(self, *args, last_poller_time=int(last_poller_time.timestamp()*1000))

                except Exception as err:
                    LOG.error(str(err))
                    LOG.error(traceback.format_exc())
                finally:
                    LOG.info("%s polling complete.", self.PACKAGE_NAME)
                    # set the last poller time for next cycle
                    last_poller_time = poller_start

                    # sleep before the next poller execution
                    exit_event.wait(getattr(self, named_poller_interval))
            exit_event.set() # loop complete

        return wrapped
    return poller_wrapper

class SOARCommon():
    """
    Common methods for accessing IBM SOAR cases and their entities: comments, attachments, etc.
    This class and its methods should be used in conjunction with wrapped
    polling logic to communicated with the SOAR platform.
    """
    def __init__(self, rest_client):
        self.rest_client = rest_client

    def _build_search_query(self, search_fields, open_cases=True):
        """[Build the json structure needed to search for cases]

        Args:
            search_fields ([dict/list]): [key/value pairs to search custom fields with specific values. If
                                         a value contains "*" then a search is used with 'has_a_value']
            NOTE: search_fields works on custom fields

        Returns:
            query_string ([dict]): [json stucture used for cases searching]
        """
        query = {
            "filters": [{
                "conditions": [
                 ]
            }],
            "sorts": [{
                "field_name": "create_date",
                "type": "desc"
            }]
        }

        if open_cases:
            field_search = {
                            "field_name": "plan_status",
                            "method": "equals",
                            "value": "A"
                          }
            query['filters'][0]['conditions'].append(field_search)

        if isinstance(search_fields, dict):
            for search_field, search_value in search_fields.items():
                field_search = {
                            "field_name": "properties.{0}".format(search_field)
                        }
                if isinstance(search_value, bool):
                    field_search['method'] = "has_a_value" if search_value else "does_not_have_a_value"
                else:
                    field_search['method'] = "equals"
                    field_search['value'] = search_value

                query['filters'][0]['conditions'].append(field_search)

        return query

    def _query_cases(self, query, filters=DEFAULT_CASES_QUERY_FILTER):
        """ run a query to find case(s) which match the query string
        Args:
            query [str]: query string to find cases
            filters [str]: Filters for the end of the uri
        Returns:
            query_results [list]: List of query results
        """
        query_uri = "/".join([INCIDENTS_URI, "query?{}".format(filters)])
        try:
            return self.rest_client.post(query_uri, query), None
        except SimpleHTTPException as err:
            LOG.error(str(err))
            LOG.error(query)
            return None, str(err)

    def _patch_case(self, case_id, case_payload):
        """
        _patch_case will update an case with the specified json payload.

        :param case_id: case ID of case to be updated.
        :param case_payload: case fields to be updated.
        :return:
        """
        try:
            # Update case
            case_url = "/".join([INCIDENTS_URI, str(case_id)])
            case = self.rest_client.get(case_url)
            patch = Patch(case)

            # Iterate over payload dict.
            for name, _ in case_payload.items():
                if name == 'properties':
                    for field_name, field_value in case_payload['properties'].items():
                        patch.add_value(field_name, field_value)
                else:
                    payload_value = case_payload.get(name)
                    patch.add_value(name, payload_value)

            patch_result = self.rest_client.patch(case_url, patch)
            result = self._chk_status(patch_result)
            # add back the case id
            result['id'] = case_id
            return result

        except Exception as err:
            LOG.error(str(err))
            raise_from(IntegrationError("_patch_case failed to update case in SOAR"), err)

    def _get_artifact_types(self):
        """ get all artifact types labels """
        return self._get_type("artifact", "type")

    def _get_incident_types(self):
        """ get all case_type_id labels """
        return self._get_type("incident", "incident_type_ids")

    def _get_resolution_types(self):
        """ get case resolution_types labels """
        return self._get_type("incident", "resolution_id")

    def _get_type(self, obj_type, lookup_field):
        """ get a specified SOAR field set of values, based on their ID (ex. status, incident_type_ids) """
        type_info = self._get_types(obj_type)

        # create a lookup table based on field name
        return { type['value']: type['label'] for type in type_info['fields'][lookup_field]['values'] }

    @cached(cache=LRUCache(maxsize=100))
    def _get_types(self, res_type):
        """ cached API call to get types information for a given type: case, artifact, etc. """
        uri = "/".join([TYPES_URI, res_type])
        return self.rest_client.get(uri)

    def _get_case_info(self, case_id, child_uri):
        """ API call for a given case and it's child objects: tasks, notes, attachments, etc. """
        try:
            uri = "/".join([INCIDENTS_URI, str(case_id)])
            if child_uri:
                uri = "/".join([uri, child_uri])

            response = self.rest_client.get(uri=uri)
            return response

        except Exception as err:
            raise_from(IntegrationError("failed to get case info from SOAR"), err)

    def _filter_comments(self, soar_comment_list, entity_comments, filter_soar_header=None):
        """
        Need to avoid creating same IBM SOAR case comments over and over
            this logic will read all SOAR comments from a case
            and remove those comments which have already sync using soar_header as a filter

        Args:
            soar_comment_list ([list]): [list of text portion of a case's comments]
            entity_comments ([list]): [list comments which the endpoint's entity contains.
                                       This will be a mix of comments sync'd from SOAR and new comments]
            filter_soar_header ([str]): [title added to SOAR comments to be filtered]
        Returns:
            new_comments ([list])
        """

        if filter_soar_header:
            # Filter entity comments with our SOAR header
            staged_entity_comments = [comment for comment in entity_comments \
                                        if clean_html(filter_soar_header) not in clean_html(comment)]
        else:
            staged_entity_comments = entity_comments.copy()

        # filter out the comments already sync'd to SOAR
        if soar_comment_list:
            already_synced = [clean_html(soar_comment) for soar_comment in soar_comment_list]
            new_entity_comments = [comment for comment in staged_entity_comments\
                if clean_html(comment) not in already_synced]
        else:
            new_entity_comments = staged_entity_comments

        return new_entity_comments

    def _chk_status(self, resp, rc=200):
        """
        Check the return status. If return code is not met, raise IntegrationError,
        if success, return the json payload

        :param resp: requests response object
        :param rc: acceptable status code or a list of two numbers indicating an acceptable range
        :type rc: int|[int, int]
        :return: json value of the resp object if call was successful
        """
        if hasattr(resp, "status_code"):
            if isinstance(rc, list):
                rc.sort()
                if resp.status_code < rc[0] or resp.status_code > rc[-1]:
                    raise IntegrationError("status code failure: {0}".format(resp.status_code))
            elif resp.status_code != rc:
                raise IntegrationError("status code failure: {0}".format(resp.status_code))

            return resp.json()

        return {}

    def get_soar_case(self, search_fields, open_cases=True, uri_filters=DEFAULT_CASES_QUERY_FILTER):
        """
        Find a SOAR case which contains custom field(s) associated with the associated endpoint.
        Returns only one case. See :class:`SOARCommon.get_soar_cases()` for examples.

        .. note::

            ``search_fields`` only supports custom fields.

        :param search_fields: Dictionary containing key/value pairs to search for a case match.
            Field values can be True/False for ``has_a_value`` or ``does_not_have_a_value``,
            otherwise a field will use ``equals`` for the value.
        :type search_fields: dict
        :param uri_filters: Filters for the end of the uri. Default is "return_level=normal"
        :type uri_filters: str
        :param open_cases: True if only querying open cases.
        :type open_cases: bool
        :return: A tuple with the matching case, if any, and any associated error message if something went wrong.
            Returns ``None`` if no associated case was found.
        :rtype: tuple(dict, str)
        """
        r_cases, error_msg = self.get_soar_cases(search_fields, open_cases=open_cases, uri_filters=uri_filters)
        if error_msg:
            return None, error_msg
        # return first case
        return (r_cases[0] if r_cases else None, None)

    def get_soar_cases(self, search_fields, open_cases=True, uri_filters=DEFAULT_CASES_QUERY_FILTER):
        """
        Get all IBM SOAR cases that match the given search fields.
        To find all cases that are synced from the endpoint platform,
        provide the unique search field that matches the unique id of your
        endpoint solution.

        **Example:**

        .. code-block:: python

            from resilient_lib import SOARCommon

            soar_common = SOARCommon(res_client)
            found_id = get_id_from_endpoint_query_result()
            cases = soar_common.get_soar_cases({ "endpoint_id": found_id }, open_cases=False)

        .. note::

            ``search_fields`` only supports custom fields.
        :param search_fields: Dictionary containing key/value pairs to search for a case match.
            Field values can be True/False for ``has_a_value`` or ``does_not_have_a_value``,
            otherwise a field will use ``equals`` for the value.
        :type search_fields: dict
        :param uri_filters: Filters for the end of the uri. Default is "return_level=normal"
        :type uri_filters: str
        :param open_cases: True if only querying open cases.
        :type open_cases: bool

        :return: A tuple with a list of cases whose values match the ``search_fields`` and any associated error message.
        :rtype: tuple(dict, str)
        """
        query = self._build_search_query(search_fields, open_cases=open_cases)
        return self._query_cases(query, filters=uri_filters)

    def create_soar_case(self, case_payload):
        """
        Create a new IBM SOAR case based on a payload formatted for the API call.

        :param case_payload: Fields to use for creating a SOAR case
        :type case_payload: dict

        :return: API result with the created case
        :rtype: dict
        """
        try:
            # Post case to IBM SOAR
            case = self.rest_client.post(INCIDENTS_URI, case_payload)
            return case
        except Exception as err:
            LOG.error(str(err))
            raise_from(IntegrationError("create_soar_case failed to create case in SOAR"), err)

    def update_soar_case(self, case_id, case_payload):
        """
        Update an IBM SOAR case (usually from a rendered Jinja template).

        :param case_payload: Fields to be updated and their values
        :type case_payload: dict
        :return: The updated SOAR case case
        :rtype: dict
        """
        try:
            result = self._patch_case(case_id, case_payload)
            return result

        except Exception as err:
            LOG.error(str(err))
            raise_from(IntegrationError("update_soar_case failed to update case in SOAR"), err)

    def create_case_comment(self, case_id, note, entity_comment_id=None, entity_comment_header=None):
        """
        Add a comment to the specified SOAR case by ID.

        :param case_id: SOAR case ID
        :type case_id: str|int
        :param note: Content to be added as a note to the case
        :type note: str
        :param entity_comment_id: (Optional) entity comment id if updating an existing comment
        :type entity_comment_id: str|int
        :param entity_comment_header: (Optional) header to place in bold at the top of the note
        :type entity_comment_header: str
        :return: Response from posting the comment to SOAR
        :rtype: dict
        """
        try:
            uri = "/".join([INCIDENTS_URI, str(case_id), "comments"])
            if entity_comment_header:
                comment = u"<b>{} ({}):</b><br>{}".format(entity_comment_header, entity_comment_id, note)
            else:
                comment = note

            note_json = {
                'format': 'html',
                'content': comment
            }
            payload = {'text': note_json}

            return self.rest_client.post(uri=uri, payload=payload)

        except Exception as err:
            LOG.error(str(err))
            raise_from(IntegrationError("failed to create comment in SOAR"), err)

    def create_datatable_row(self, case_id, datatable, rowdata):
        """
        Create a row in a SOAR datatable.
        ``rowdata`` should be formatted as a dictionary of
        column name and value pairs

        **Example:**

        .. code-block:: python

            from resilient_lib import SOARCommon

            soar_common = SOARCommon(res_client)

            case_id = get_case_id()
            rowdata = {"column_1_api_name": 1, "column_2_api_name": 2}

            soar_common.create_datatable_row(case_id, "my_dt", rowdata)

        :param case_id: case containing the datatable
        :type case_id: str|int
        :param datatable: API name of datatable
        :type datatable: str
        :param rowdata: columns and values to add
        :type rowdata: dict
        """
        uri = "/".join([INCIDENTS_URI, str(case_id), "table_data", datatable, "row_data"])


        formatted_cells = {}

        for column in rowdata:
            formatted_cells[column] = {"value": rowdata.get(column)}

        formatted_cells = {"cells": formatted_cells}

        return self.rest_client.post(uri=uri, payload=formatted_cells)

    def get_case(self, case_id):
        """
        Get a SOAR case based on the case id
        
        :param case_id: ID of the case to get the details of
        :type case_id: str|int
        :return: Details of the case for a given ID
        :rtype: dict
        """
        case = self._get_case_info(case_id, None)
        # create a new field of the incident_type_ids in label form
        case['case_types'] = self.lookup_incident_types(case['incident_type_ids'])
        return case

    def get_case_attachment(self, case_id, artifact_id=None, task_id=None, attachment_id=None, return_base64=True):
        """
        Get contents of a file attachment or artifact.

        * If ``incident_id`` and ``artifact_id`` are defined it will get that Artifact
        * If ``incident_id`` and ``attachment_id`` are defined it will get that Incident Attachment
        * If ``incident_id``, ``task_id`` and ``attachment_id`` are defined it will get that Task Attachment
        
        :param case_id: ID of the case to get the details of
        :type case_id: str|int
        :param artifact_id: ID of the Incident's Artifact to download
        :type artifact_id: str|int
        :param task_id: ID of the Task to download it's Attachment from
        :type task_id: str|int
        :param attachment_id: id of the Incident's Attachment to download
        :type attachment_id: int|str
        :param return_base64: if False, return value of the content will be given as bytes, default True returns a base64 encoded string
        :type return_base64: bool
        :return: name of the artifact or file and base64 encoded string or byte string of attachment
        :rtype: tuple(str, bytes|base64(str))
        """
        file_content = get_file_attachment(self.rest_client, case_id, artifact_id=artifact_id, task_id=task_id, attachment_id=attachment_id)
        if return_base64:
            file_content = b_to_s(base64.b64encode(file_content))

        file_name = get_file_attachment_name(self.rest_client, case_id, artifact_id=artifact_id, task_id=task_id, attachment_id=attachment_id)
        return file_name, file_content

    def get_case_artifacts(self, case_id):
        """
        Get all case artifacts

        :param case_id: ID of the case to get the details of
        :type case_id: str|int
        """
        return self._get_case_info(case_id, "artifacts")

    def get_case_comments(self, case_id):
        """
        Get all case comments 

        :param case_id: ID of the case to get the details of
        :type case_id: str|int
        """
        return self._get_case_info(case_id, "comments")

    def get_case_attachments(self, case_id, return_base64=True):
        """
        Get all case attachments
        
        :param case_id: ID of the case to get the details of
        :type case_id: str|int
        :param return_base64: if False, return value of the content will be given as bytes
        :type return_base64: bool
        :return: list of attachments associated with the given case
        :rtype: list[tuple(str, bytes|base64(str))]
        """
        attachments =  self._get_case_info(case_id, "attachments")
        for attachment in attachments:
            _, attachment['content'] = self.get_case_attachment(case_id, attachment_id=attachment['id'], return_base64=return_base64)

        return attachments

    def lookup_artifact_type(self, artifact_type):
        """
        Return an artifact type based on it's ID on SOAR. If not found, return None

        :param artifact_type: artifact type to search for in SOAR
        :type artifact_type: str
        :return: found artifact type if found in SOAR
        :rtype: str
        """
        try:
            types = self._get_artifact_types()
            if artifact_type in types:
                return types[artifact_type]
        except SimpleHTTPException:
            pass
        return None

    def lookup_incident_types(self, incident_type_ids):
        """
        Return an incident type based on it's ID. If not found, return None

        :param incident_type_ids: list of incident type IDs to check for in SOAR
        :type incident_type_ids: list(str)
        :return: list of incident types that match in SOAR what was given
        :rtype: list(str)
        """
        if not incident_type_ids:
            return incident_type_ids

        types = self._get_incident_types()

        return [types[incident_type] for incident_type in incident_type_ids if incident_type in types]

    def filter_soar_comments(self, case_id, entity_comments, soar_header=None):
        """
        Read all SOAR comments from a case and remove those comments which have
        already been synced using ``soar_header`` as a filter

        :param case_id: ([str]): [IBM SOAR case id]
        :type case_id: str|int
        :param entity_comments: list comments which the endpoint's entity contains.
            This will be a mix of comments sync'd from SOAR and new comments
        :type entity_comments: list(str)
        :param soar_header: title added to SOAR comments to be filtered
        :type soar_header: str
        :return: list of remaining comments
        :rtype: list
        """
        # get the SOAR case comments and capture only the text
        soar_comments = self.get_case_comments(str(case_id))
        soar_comment_list = [comment['text'] for comment in soar_comments]

        # filter entity comments with our SOAR header
        return self._filter_comments(soar_comment_list, entity_comments, filter_soar_header=soar_header)


@cached(cache=LRUCache(maxsize=100))
def eval_mapping(eval_value, wrapper=None):
    """
    Map a JSON string to a python object. Safely evaluates the values with the use
    of ``ast.literal_eval`` and can ONLY convert to a list or a dictionary.

    .. note::

        ``wrapper`` is optional (but recommended) to ensure that the ``eval_value``
        is properly formatted as JSON. If ``None``, then assumes
        that the ``eval_value`` comes pre-formatted in acceptable JSON format

    This method is intended to be used to safely take input strings from the user's
    config file that will eventually need to evaluated as JSON elsewhere in the code.

    **Example:**

    .. code-block:: python

        from resilient_lib import eval_mapping

        config_value = app_configs.get("some_config")
        # Assume that 'config_value' has a value of
        # '"source":"A","tags":["tagA"],"priorities":[40,50]'

        # parse the values to a dict, using a dictionary wrapper:
        parsed_config = eval_mapping(config_value, "{{ {} }}")

        # the returned value is a dict like:
        # {
        #   "source": "A",
        #   "tags": ["tagA"],
        #   "priorities": [40, 50]
        # }

    :param eval_value: json fragment to evaluate
    :type eval_value: str
    :param wrapper: wrapper to put the source value into ``[{}]`` or ``{{ {} }}``
    :type wrapper: str
    :return: converted data
    :rtype: list|dict
    """
    if not eval_value:
        return None

    try:
        if wrapper:
            eval_value = wrapper.format(eval_value)

        # Try converting input to a dict or array
        result = literal_eval(eval_value)
        # expecting the result to be a list or dictionary
        if not isinstance(result, (list, dict)):
            raise IntegrationError("Unexpected result: %s", result)

        return result

    except Exception as err:
        LOG.error(str(err))
        LOG.error(traceback.format_exc())
        LOG.error("""mapping eval_value must be a string representation of a (partial) array or dictionary e.g. "'value1', 'value2'" or "'key':'value'" """)

    return None

def s_to_b(value):
    """ string to bytes """
    try:
        return bytes(value, 'utf-8')
    except Exception:
        return value

def b_to_s(value):
    """ bytes to string """
    try:
        return value.decode()
    except Exception:
        return value

def get_last_poller_date(polling_lookback):
    """
    Get the last poller datetime based on a lookback value
    Args:
        polling_lookback ([number]): # of minutes to lookback
    Returns:
        [datetime]: [datetime to use for last poller run time]
    """
    return _get_timestamp() - datetime.timedelta(minutes=polling_lookback)

def _get_timestamp():
    """
    Get the current timestamp

    Returns:
        datetime: current datetime
    """
    return datetime.datetime.now()



import pytest
from mock import patch
from resilient_lib.ui import Field, Tab
import uuid

class TestConditions(object):
    def test_can_initialize(self):
        assert Field('test').conditions is not None
        assert Field('test').conditions.has_value() is not None
        assert Field('test').conditions.equals("test") is not None
    
    def test_conditions_added_to_dto(self):
        class TestTab(Tab):
            NAME = "test_tab"
            UUID = uuid.uuid4()
            SECTION = "fn_test"
            
            CONTAINS = [Field("test")]
            SHOW_IF = [Field("test").conditions.contains("value")]
        
        assert TestTab.as_dto()['show_if'] == TestTab.SHOW_IF
